"""WIP example using `minGPT` with DistributedDataParallel on both CPU and JAX.

Required `mingpt` package for model definition (see requirements.txt). Some
hyperparameters and training configuration borrowed from nanoGPT:
https://github.com/karpathy/nanoGPT

Example command (single host):
torchrun --standalone xla/experimental/torchax/examples/train_gpt/train_ddp.py

Tested on a TPU v4-8
"""

import datetime
import jax
import torch
import torch.utils.data
import torch.utils.data.distributed
import torch.distributed as dist
import torch.optim as optim
import torchax
from tqdm import tqdm
from mingpt.model import GPT
from datasets import load_dataset
import tiktoken
import pathlib
import torch.utils._pytree as torch_pytree


def _checkpoint(jax_model, path: pathlib.Path):
  torch.save(
      torch_pytree.tree_map_only(
          torchax.tensor.Tensor,
          torchax.tensor.Tensor.torch,
          jax_model.state_dict(),
      ),
      path,
  )


def main():
  dist.init_process_group(backend="gloo")
  dataset_name = "Skylion007/openwebtext"
  dataset = load_dataset(dataset_name, split="train", trust_remote_code=True)

  enc = tiktoken.get_encoding("gpt2")

  def tokenize(ex):
    """Tokenize each example and append the end-of-text token."""
    ids = enc.encode_ordinary(ex["text"])
    ids.append(enc.eot_token)
    return {"ids": ids}

  dataset = dataset.map(tokenize, num_proc=16)

  def group_texts(exs):
    """Group batches of tokens into `block_size` chunks."""
    cat = torch.cat([torch.tensor(ex) for ex in exs["ids"]])
    total_len = cat.size()[0]
    num_chunks = total_len // 1025
    split = torch.split(cat[:num_chunks * 1025], 1025)
    xs = [ex[:-1] for ex in split]
    ys = [ex[1:] for ex in split]
    return {"x": xs, "y": ys}

  dataset = dataset.map(
      group_texts, batched=True, remove_columns=["text", "ids"], num_proc=16)
  dataset.shard(dist.get_world_size(), dist.get_rank())
  env = torchax.default_env()

  print(jax.device_count(), "devices")

  torch.manual_seed(0)
  per_device_batch_size = 8
  local_batch_size = jax.local_device_count() * per_device_batch_size
  global_batch_size = jax.device_count() * per_device_batch_size
  dataloader = torch.utils.data.DataLoader(
      dataset.with_format("torch"), batch_size=local_batch_size, drop_last=True)

  # Create model and wrap with DDP
  def create_model():
    torch.manual_seed(0)
    model_config = GPT.get_default_config()
    model_config.model_type = "gpt2"
    model_config.vocab_size = enc.n_vocab
    model_config.block_size = 1024
    # TODO: use bf16 when erroneous type promotions are fixed
    return GPT(model_config)  # .to(dtype=torch.bfloat16)

  checkpoint_subdir = pathlib.Path(
      "checkpoints") / datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
  checkpoint_subdir.mkdir(parents=True)
  jax_model = torchax.distributed.DistributedDataParallel(create_model(), env)

  # TODO: LR scheduler
  jax_optimizer = optim.SGD(jax_model.parameters(), lr=6e-4, weight_decay=0.1)

  # Contents of `step_fn` can be inlined if using eager
  @jax_model.jit_step
  def step_fn(jax_data, jax_target):
    jax_optimizer.zero_grad()
    jax_output, jax_loss = jax_model(jax_data, jax_target)
    jax_loss.backward()
    torch.nn.utils.clip_grad_norm_(jax_model.parameters(), 1.0)
    jax_optimizer.step()

    return jax_output, jax_loss

  tokens_per_batch = global_batch_size * 1024

  for epoch in range(1):
    print("epoch", epoch)
    for i, batch in enumerate(
        tqdm(dataloader, unit="tok", unit_scale=tokens_per_batch)):
      data, target = batch["x"], batch["y"]
      jax_data, jax_target = env.j2t_iso(
          (jax_model.shard_input(data), jax_model.shard_input(target)))
      jax_output, jax_loss = step_fn(jax_data, jax_target)

      if i % 1000 == 0:
        _checkpoint(jax_model, checkpoint_subdir / "gpt2_124m_{epoch}_{i}.ckpt")
        print("step", i, jax_loss.item())

  with torch.no_grad():
    with env:
      inp = enc.encode("This GPT-2 example is")
      input_jax = torch.tensor([inp], dtype=torch.long)
      # TODO: need to access underlying module for methods
      jax_generated = jax_model._module.generate(
          jax_model.replicate_input(input_jax),
          100,
          do_sample=False,
      )

  print("input sequence:", inp, enc.decode(inp))
  print(jax_generated)
  print("predicted (JAX):", enc.decode(jax_generated.numpy().tolist()))


if __name__ == "__main__":
  main()



import contextlib
import enum
import torch
from torch.utils import _pytree as pytree


# enum class CastPolicy : uint8_t {
#   lower_precision_fp = 0, // Cast all inputs to lower_precision_fp before
#                           // running the op. Currently, lower_precision_fp is
#                           // fp16 for AutocastCUDA, and is defined by user
#                           // (default bf16) for AutocastCPU or other device.
#   fp32, // Cast all inputs to at::kFloat before running the op.
#   fp32_set_opt_dtype, // Treats functions (like softmax) that
#                       //  1. we'd like to run in fp32 and
#                       //  2. have a std::optional<ScalarType> arg that controls
#                       //  the output type.
#                       // fp32_set_opt_dtype wrappers' policy is: if the output
#                       // type is already set, don't touch it, otherwise, set
#                       // it to at::kFloat.
#   fp32_append_dtype, // Treats functions (like norm) that
#                      //  1. we'd like to run in fp32 and
#                      //  2. have some overloads that accept an output type and
#                      //  other overloads that don't.
#                      // fp32_append_dtype wrappers wrap the overloads that don't
#                      // have an output dtype.
#                      // The wrapper policy is:  append at::kFloat to the args,
#                      // and redispatch to the type-aware overload.
#   promote, // Run in the widest dtype among several args.
# };
class CastPolicy(enum.Enum):
  LOWER_PRECISION_FP = 0
  FP32 = 1
  FP32_SET_OPT_DTYPE = 2
  FP32_APPEND_DTYPE = 3
  PROMOTE = 4


def execute_policy(policy, args, kwargs, target_lower_fp):

  def is_float(a):
    return isinstance(a, torch.Tensor) and a.is_floating_point()
  match policy:
    case CastPolicy.LOWER_PRECISION_FP:
      return pytree.tree_map_only(is_float, lambda a: a.to(target_lower_fp),
                                  (args, kwargs))
    case CastPolicy.FP32:
      return pytree.tree_map_only(is_float, lambda a: a.to(torch.float32),
                                  (args, kwargs))
    case CastPolicy.PROMOTE:
      dtypes = set(a.dtype for a in args)
      widest = max((dtype.itemsize, dtype) for dtype in dtypes)[1]
      return pytree.tree_map_only(is_float, lambda a: a.to(widest),
                                  (args, kwargs))
    case _:
      raise AssertionError(f'Policy {policy} not implemented yet.')


@contextlib.contextmanager
def autocast(device, dtype=torch.bfloat16, env=None):
  del device
  if env is None:
    import torchax
    env = torchax.default_env()
  env.autocast_dtype, old = dtype, env.autocast_dtype
  yield
  env.autocast_dtype = old


# https://github.com/pytorch/pytorch/blob/05faba40287cf7d8734da96cb2e904f39710bf29/aten/src/ATen/autocast_mode.cpp#L327
autocast_policy = {
    torch.ops.aten.conv1d.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv1d.padding:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv2d.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv2d.padding:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv3d.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv3d.padding:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.bmm.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.mm.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.linalg_vecdot.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.baddbmm.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.addmm.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten._addmm_activation.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.addbmm.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.linear.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten._convolution.deprecated:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.matmul.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv_tbc.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.mkldnn_rnn_layer.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv_transpose1d.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv_transpose2d.input:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.conv_transpose3d.input:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.prelu.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten.scaled_dot_product_attention.default:
        CastPolicy.LOWER_PRECISION_FP,
    torch.ops.aten._native_multi_head_attention.default:
        CastPolicy.LOWER_PRECISION_FP,

    # fp32 cast policy
    torch.ops.aten.avg_pool3d.default:
        CastPolicy.FP32,
    torch.ops.aten.binary_cross_entropy.default:
        CastPolicy.FP32,
    torch.ops.aten.grid_sampler.default:
        CastPolicy.FP32,
    torch.ops.aten.polar.default:
        CastPolicy.FP32,
    torch.ops.aten.prod.default:
        CastPolicy.FP32,
    torch.ops.aten.prod.dim_int:
        CastPolicy.FP32,
    torch.ops.aten.prod.dim_Dimname:
        CastPolicy.FP32,
    torch.ops.aten.quantile.default:
        CastPolicy.FP32,
    torch.ops.aten.quantile.scalar:
        CastPolicy.FP32,
    torch.ops.aten.nanquantile.default:
        CastPolicy.FP32,
    torch.ops.aten.nanquantile.scalar:
        CastPolicy.FP32,
    torch.ops.aten.stft.default:
        CastPolicy.FP32,
    torch.ops.aten.stft.center:
        CastPolicy.FP32,
    torch.ops.aten.cdist.default:
        CastPolicy.FP32,
    torch.ops.aten.grid_sampler_2d.default:
        CastPolicy.FP32,
    torch.ops.aten._grid_sampler_2d_cpu_fallback.default:
        CastPolicy.FP32,
    torch.ops.aten.grid_sampler_3d.default:
        CastPolicy.FP32,
    torch.ops.aten.trace.default:
        CastPolicy.FP32,
    torch.ops.aten.view_as_complex.default:
        CastPolicy.FP32,
    torch.ops.aten.cholesky.default:
        CastPolicy.FP32,
    torch.ops.aten.cholesky_inverse.default:
        CastPolicy.FP32,
    torch.ops.aten.cholesky_solve.default:
        CastPolicy.FP32,
    torch.ops.aten.inverse.default:
        CastPolicy.FP32,
    torch.ops.aten.lu_solve.default:
        CastPolicy.FP32,
    torch.ops.aten.orgqr.default:
        CastPolicy.FP32,
    torch.ops.aten.ormqr.default:
        CastPolicy.FP32,
    torch.ops.aten.pinverse.default:
        CastPolicy.FP32,
    torch.ops.aten.max_pool3d.default:
        CastPolicy.FP32,
    torch.ops.aten.max_unpool2d.default:
        CastPolicy.FP32,
    torch.ops.aten.max_unpool3d.default:
        CastPolicy.FP32,
    torch.ops.aten.adaptive_avg_pool3d.default:
        CastPolicy.FP32,
    torch.ops.aten.reflection_pad1d.default:
        CastPolicy.FP32,
    torch.ops.aten.reflection_pad2d.default:
        CastPolicy.FP32,
    torch.ops.aten.replication_pad1d.default:
        CastPolicy.FP32,
    torch.ops.aten.replication_pad2d.default:
        CastPolicy.FP32,
    torch.ops.aten.replication_pad3d.default:
        CastPolicy.FP32,
    torch.ops.aten.mse_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.cosine_embedding_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.nll_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.nll_loss2d.default:
        CastPolicy.FP32,
    torch.ops.aten.hinge_embedding_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.poisson_nll_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.smooth_l1_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.cross_entropy_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.l1_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.huber_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.margin_ranking_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.soft_margin_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.triplet_margin_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.multi_margin_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.ctc_loss.IntList:
        CastPolicy.FP32,
    torch.ops.aten.ctc_loss.Tensor:
        CastPolicy.FP32,
    torch.ops.aten.kl_div.default:
        CastPolicy.FP32,
    torch.ops.aten.multilabel_margin_loss.default:
        CastPolicy.FP32,
    torch.ops.aten.binary_cross_entropy_with_logits.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_fft.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_ifft.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_fft2.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_ifft2.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_fftn.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_ifftn.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_rfft.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_irfft.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_rfft2.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_irfft2.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_rfftn.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_irfftn.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_hfft.default:
        CastPolicy.FP32,
    torch.ops.aten.fft_ihfft.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_cond.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_cond.p_str:
        CastPolicy.FP32,
    torch.ops.aten.linalg_matrix_rank.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_matrix_rank.tol_tensor:
        CastPolicy.FP32,
    torch.ops.aten.linalg_matrix_rank.atol_rtol_tensor:
        CastPolicy.FP32,
    torch.ops.aten.linalg_matrix_rank.atol_rtol_float:
        CastPolicy.FP32,
    torch.ops.aten.linalg_solve.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_cholesky.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_svdvals.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_eigvals.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_eigvalsh.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_inv.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_householder_product.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_tensorinv.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_tensorsolve.default:
        CastPolicy.FP32,
    torch.ops.aten.fake_quantize_per_tensor_affine.default:
        CastPolicy.FP32,
    torch.ops.aten.geqrf.default:
        CastPolicy.FP32,
    torch.ops.aten._lu_with_info.default:
        CastPolicy.FP32,
    torch.ops.aten.qr.default:
        CastPolicy.FP32,
    torch.ops.aten.svd.default:
        CastPolicy.FP32,
    torch.ops.aten.triangular_solve.default:
        CastPolicy.FP32,
    torch.ops.aten.fractional_max_pool2d.default:
        CastPolicy.FP32,
    torch.ops.aten.fractional_max_pool3d.default:
        CastPolicy.FP32,
    torch.ops.aten.adaptive_max_pool3d.default:
        CastPolicy.FP32,
    torch.ops.aten.multilabel_margin_loss_forward.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_qr.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_cholesky_ex.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_svd.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_eig.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_eigh.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_lstsq.default:
        CastPolicy.FP32,
    torch.ops.aten.linalg_inv_ex.default:
        CastPolicy.FP32,

    # promote
    torch.ops.aten.stack.default:
        CastPolicy.PROMOTE,
    torch.ops.aten.cat.default:
        CastPolicy.PROMOTE,
    torch.ops.aten.index_copy.default:
        CastPolicy.PROMOTE,
    torch.ops.aten.index_copy.dimname:
        CastPolicy.PROMOTE,
}



import os
import jax
import numpy as np

import pytest
import torch
import torch.distributed._functional_collectives
import torch.distributed as dist
import torchax
import torchax.distributed

# Dummy group name to use with functional collectives. Ignored by
# implementations.
# TODO(wcromar): do something useful with group name
GROUP_NAME = "process_group"

torchax.enable_globally()


@pytest.fixture(scope="module")
def multi_cpu():
  # TODO(wcromar): support other device counts
  assert (jax.device_count() == 4
         ), "Set XLA_FLAGS=--xla_force_host_platform_device_count=4 if on CPU"

  yield jax.device_count()


@pytest.fixture()
def process_group():
  os.environ["MASTER_ADDR"] = "localhost"
  os.environ["MASTER_PORT"] = "12355"
  dist.init_process_group(backend="jax", init_method="jax://")
  # HACK: our default process group has world size 1, regardless of actual
  # device count. Only put rank 0 so PyTorch doesn't complain about non-existent
  # ranks. Our lowerings ignore this list, so this ends up being fine.
  # TODO(wcromar): Figure out if there's a cleaner way
  group_ranks = [0]
  yield group_ranks
  dist.destroy_process_group()


def test_all_gather_tensor(multi_cpu, process_group):
  device_count = multi_cpu

  def f(index: torchax.tensor.Tensor):
    with torchax.default_env():
      output = torch.zeros_like(index).expand(device_count)
    dist.all_gather_into_tensor(output, index)
    return output

  res = torchax.distributed.spawn(f)

  expected_tensors = [[0, 1, 2, 3] for _ in range(device_count)]
  np.testing.assert_equal([r.numpy() for r in res], expected_tensors)


def test_all_gather_tensor_func(multi_cpu, process_group):
  device_count = multi_cpu
  group_ranks = process_group

  def f(index: torchax.tensor.Tensor):
    return torch.distributed._functional_collectives.all_gather_tensor(
        index, 0, group_ranks)

  res = torchax.distributed.spawn(f)

  expected_tensors = [[0, 1, 2, 3] for _ in range(device_count)]
  np.testing.assert_equal([r.numpy() for r in res], expected_tensors)


@pytest.mark.parametrize(
    ("op", "expected"),
    [
        (dist.ReduceOp.SUM, sum(range(4))),
        (dist.ReduceOp.AVG, sum(range(4)) // 4),
        (dist.ReduceOp.MIN, 0),
        (dist.ReduceOp.MAX, 3),
    ],
)
def test_all_reduce(op, expected, multi_cpu, process_group):
  device_count = multi_cpu

  def f(index):
    with torchax.default_env():
      dist.all_reduce(index, op)
      return index

  res = torchax.distributed.spawn(f)

  expected_tensors = [expected for _ in range(device_count)]
  np.testing.assert_equal(res.numpy(), expected_tensors)


@pytest.mark.parametrize(
    ("op", "expected"),
    [
        ("sum", sum(range(4))),
        ("avg", sum(range(4)) / 4),
        ("min", 0),
        ("max", 3),
    ],
)
def test_all_reduce_func(op, expected, multi_cpu):
  device_count = multi_cpu

  def f(index):
    return torch.distributed._functional_collectives.all_reduce(
        index, op, GROUP_NAME)

  res = torchax.distributed.spawn(f)

  expected_tensors = [expected for _ in range(device_count)]
  np.testing.assert_equal(res.numpy(), expected_tensors)


@pytest.mark.parametrize(
    ("rank", "expected"),
    [
        (0, 0),
        (2, 2),
    ],
)
def test_broadcast(rank, expected, multi_cpu, process_group):
  device_count = multi_cpu

  def f(index):
    dist.broadcast(index, rank)
    return index

  res = torchax.distributed.spawn(f)

  expected_tensors = [expected for _ in range(device_count)]
  np.testing.assert_equal(res.numpy(), expected_tensors)


@pytest.mark.parametrize(
    ("rank", "expected"),
    [
        (0, 0),
        (2, 2),
    ],
)
def test_broadcast_func(rank, expected, multi_cpu):
  device_count = multi_cpu

  def f(index):
    return torch.distributed._functional_collectives.broadcast(
        index, rank, GROUP_NAME)

  res = torchax.distributed.spawn(f)

  expected_tensors = [expected for _ in range(device_count)]
  np.testing.assert_equal(res.numpy(), expected_tensors)



import collections
import functools
import torch
import jax
import torchax
from torchax import interop
from torchax.interop import torch_view, jax_view
import optax

remat = torch_view(jax.remat)
mark_sharding = torch_view(jax.lax.with_sharding_constraint)


def make_train_step(model_fn, loss_fn, optax_optimizer, remat_policy=None):
  """Make a function that do one train step given model and loss.

  model_fn: a function representing the model's forward:
      i.e. has signature Callable[weights, buffers, args] -> result. Where,
      weights is a pytree of trainable parameters
      buffers is a pytree of non-trainable parameters / constants
      args is the input data loaded from the data set
      result is the return value of the model
  loss_fn: a function to compute loss.
      i.e. it has signature of Callable[result, label] -> loss
      where, result is what model_fn returned
        loss is loaded from the dataloader.
  optax_optimizer: the optimizer from optax library. for example, optax.adam
  remat_policy: One of jax.ad_checkpoint.checkpoint_policies, specifies how
      to do gradient checkpointing. If None, then it means checkpoint everything.
  """
  env = torchax.default_env()

  def loss(weights, buffers, args, label):  # inputs are XLATensor
    with env, jax.named_scope('compute_loss'):
      res = model_fn(weights, buffers, args)
      l = loss_fn(res, label)
      return l

  loss = interop.gradient_checkpoint(loss, kwargs={'policy': remat_policy})
  grad_fn = interop.jax_value_and_grad(loss)

  def step(weights, buffers, opt_state, args, label):  #inputs are array
    with jax.named_scope('compute_gradient'):
      loss, gradient = grad_fn(weights, buffers, args, label)

    with jax.named_scope("optimizer_updates"):
      updates, opt_state = interop.call_jax(optax_optimizer.update, gradient,
                                            opt_state, weights)
      weights = interop.call_jax(optax.apply_updates, weights, updates)
    return loss, weights, opt_state

  # TODO: apply jax.jit so the user don't have to.
  return step


class Container:
  pass


class ScannedModule(torch.nn.Module):

  def __init__(self, module_list, checkpoint_policy=None):
    super().__init__()

    self.c = None
    assert module_list
    self.c = Container()
    self.c.one_mod = module_list[0]
    self.checkpoint_policy = checkpoint_policy

    weights = self._stack_layer_weights(module_list)
    self.layer_weights_keys = list(self.c.one_mod.state_dict().keys())
    self.params = torch.nn.ParameterDict({
        self._param_name_new(k): v for k, v in weights.items()
    })

  def _stack_layer_weights(self, module_list):
    # Create weights such that, for every [n, m] weights
    # becomes [k, n, m] where k is number of layer
    # i.e. stacking layer weights together
    temp = collections.defaultdict(list)
    for m in module_list:
      for k, v in m.state_dict().items():
        temp[k].append(v)
    res = {k: torch.stack(v) for k, v in temp.items()}
    return res

  def _param_name_new(self, old):
    return '___'.join(old.split('.'))

  def _param_name_old(self, new):
    return '.'.join(new.split('___'))

  def forward(self, *args, **kwargs):
    assert not kwargs
    weights = {
        k: self.params[self._param_name_new(k)] for k in self.layer_weights_keys
    }
    scan = interop.torch_view(jax.lax.scan)

    def eval_one_layer(args, weight):
      # unpack args
      h, *rest = args
      newh = torch.func.functional_call(self.c.one_mod, weight, args)
      # next layer's input; and residual to be added to list
      return (newh, *rest), None

    _eval_one_layer = interop.gradient_checkpoint(
        eval_one_layer,
        kwargs={'policy': self.checkpoint_policy},
    )
    h, _ = scan(
        _eval_one_layer,
        args,
        weights,
    )
    return h[0]



"""The `jlibrary` module has functions which help to preserve torch.library ops
during export. This includes aten ops, and custom operations.
"""

import torch
import torch.nn as nn
import torchax
from torchax.ops import jaten
import jax
import functools


def _jit_composite_impl(composite_name, jaxpr_impl, **jit_args):
  """Wrap a jaxpr in a jitted function with the proper composite name
  TODO: Wrap JIT in a `stablehlo.composite` op, instead of generating a call op.
  """

  def composite_impl(*args):
    return jaxpr_impl(*args)

  composite_impl.__name__ = composite_name
  composite_impl.__qualname__ = composite_name
  return jax.jit(composite_impl, **jit_args)


def register_jax_composite(composite_name, impl, *ops, **jit_args):
  """Register a composite using a JAX implementation.
    composite_name - The name of the library op to use in the exported composite
    impl           - A JAX lowering for the library operation
    *ops           - Variadic torch.ops to lower using `impl`.
    **jit_args     - Additional parameters to forward to JAX jit.

  This is used to register custom lowerings with an explicit jaxpr
  implementation, such as preserving a specific aten op using a jaten impl.

  For custom torch op registration with a decomposition written in torch,
  use `register_torch_composite`.

  For jit params and troubleshooting see:
  https://jax.readthedocs.io/en/latest/_autosummary/jax.jit.html
  """

  @jaten.op(*ops)
  def _composite_impl(*args):
    return _jit_composite_impl(composite_name, impl, **jit_args)(*args)


def register_torch_composite(composite_name, impl, *ops, **jit_args):
  """Register a torch decomposition as a composite.
  This is useful for registerring custom torch op libraries as composite ops.

  The `impl` can be the `@impl` used to define the torch custom library op.
  This must be a function or module impl that provides the decompositions, and
  not an instance of the custom op.

  TODO: Better error handling, or can we make this an instance of the op as a param?

  For jit params and troubleshooting see:
  https://jax.readthedocs.io/en/latest/_autosummary/jax.jit.html
  """

  @jaten.op(*ops)
  def _composite_impl(*args):

    class ImplWrapper(torch.nn.Module):

      def __init__(self):
        super().__init__()

      def forward(self, *args):
        return impl(*args)

    # Note: avoid refactoring to share code with register_jaxpr_composite.
    # The `extract_jax` call must live in the `@jaten.op` handler. If called
    # outside of the handler, we would build the jaxpr representation of the
    # module once during registration, potentially missing op registrations that
    # come after. I.e. may miss nested abstractions if we build jaxpr AoT.
    state, jfn = torchax.extract_jax(ImplWrapper())
    jaxpr_impl = lambda *args: jfn(state, tuple([*args]))
    return _jit_composite_impl(composite_name, jaxpr_impl, **jit_args)(*args)



import torchax
from torch import nn
from torch.nn import functional as F
import torch

xla_env = torchax.enable_globally()


class MyModel(nn.Module):

  def __init__(self):
    super().__init__()
    self.fc1 = nn.Linear(28 * 28, 120)
    self.fc2 = nn.Linear(120, 84)
    self.fc3 = nn.Linear(84, 10)

  def forward(self, x):
    x = x.view(-1, 28 * 28)
    x = F.relu(self.fc1(x))
    x = F.relu(self.fc2(x))
    x = self.fc3(x)
    return x


m = MyModel()
m = m.to('jax')

# Execute this model using torch
inputs = torch.randn(3, 3, 28, 28, device='jax')

print(m(inputs))
print('---=====')

m_compiled = torchax.compile(m)

print(m_compiled(inputs))

print('---')



"""Tensor constructor overrides"""

import math
import collections.abc
import functools
from typing import Optional, Sequence, Tuple
import numpy as np

import jax
import jax.numpy as jnp
from jax.experimental.pallas.ops.tpu import flash_attention
from jax.experimental.shard_map import shard_map

import torch
from torchax.ops.ops_registry import register_torch_function_op
from torchax.ops import op_base, mappings, jaten, jimage
import torchax.tensor
from torchax.view import View, NarrowInfo
import torch.utils._pytree as pytree


def register_function(torch_func, **kwargs):
  return functools.partial(register_torch_function_op, torch_func, **kwargs)


@register_function(torch.as_tensor, is_jax_function=False, needs_env=True)
@op_base.convert_dtype(
    use_default_dtype=False)  # Attempt to infer type from elements
def _as_tensor(data, dtype=None, device=None, env=None):
  if isinstance(data, torch.Tensor):
    return env._to_copy(data, dtype, device)
  if isinstance(data, np.ndarray):
    jax_res = jnp.asarray(data)
  else:
    jax_res = _tensor(data, dtype=dtype)
  return torchax.tensor.Tensor(jax_res, env)


@register_function(torch.tensor)
@op_base.convert_dtype(
    use_default_dtype=False)  # Attempt to infer type from elements
def _tensor(data, *, dtype=None, **kwargs):
  python_types_to_torch_types = {
      bool: jnp.bool,
      int: jnp.int64,
      float: jnp.float32,
      complex: jnp.complex64,
  }
  if not dtype:
    leaves = jax.tree_util.tree_leaves(data)
    if len(leaves) > 0:
      dtype = python_types_to_torch_types.get(type(leaves[0]))

  return jnp.array(
      data, dtype=dtype or mappings.t2j_dtype(torch.get_default_dtype()))


@register_function(torch.allclose)
def _aten_allclose(input, other, rtol=1e-05, atol=1e-08, equal_nan=False):
  return jnp.allclose(input, other, rtol, atol, equal_nan)


@register_function(torch.angle)
def _torch_angle(input):
  if input.dtype.name == "int64":
    input = input.astype(jnp.dtype("float32"))
  return jnp.angle(input)


@register_function(torch.argsort)
def _torch_argsort(input, dim=-1, descending=False, stable=False):
  expanded = False
  if input.ndim == 0:
    # for self of rank 0:
    # torch.any(x, 0), torch.any(x, -1) works;
    # torch.any(x, 1) throws out of bounds, so it's
    # behavior is the same as a jnp array of rank 1
    expanded = True
    input = jnp.expand_dims(input, 0)
  res = jnp.argsort(input, axis=dim, descending=descending, stable=stable)
  if expanded:
    res = res.squeeze()
  return res


@register_function(torch.diag)
def _diag(input, diagonal=0):
  return jnp.diag(input, k=diagonal)


@register_function(torch.einsum)
@register_function(torch.ops.aten.einsum)
def _einsum(equation, *operands):

  def get_params(*a):
    inner_list = a[0]
    if not isinstance(inner_list, jax.Array):
      if len(inner_list) == 1:
        A = inner_list
        return A
      elif len(inner_list) == 2:
        A, B = inner_list
        return A, B
    return operands

  assert isinstance(equation, str), "Only accept str equation"
  filtered_operands = get_params(*operands)
  return jnp.einsum(equation, *filtered_operands)


def _sdpa_reference(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    scale=None,
    enable_gqa=False,
) -> torch.Tensor:
  L, S = query.size(-2), key.size(-2)
  scale_factor = 1 / math.sqrt(query.size(-1)) if scale is None else scale
  attn_bias = torch.zeros(L, S, dtype=query.dtype, device=query.device)
  if is_causal:
    assert attn_mask is None
    temp_mask = torch.ones(
        L, S, dtype=torch.bool, device=query.device).tril(diagonal=0)
    attn_bias.masked_fill_(temp_mask.logical_not(), float("-inf"))
    attn_bias.to(query.dtype)
  if attn_mask is not None:
    if attn_mask.dtype == torch.bool:
      attn_bias.masked_fill_(attn_mask.logical_not(), float("-inf"))
    else:
      attn_bias += attn_mask
  if enable_gqa:
    key = key.repeat_interleave(query.size(-3) // key.size(-3), -3)
    value = value.repeat_interleave(query.size(-3) // value.size(-3), -3)

  attn_weight = query @ key.transpose(-2, -1) * scale_factor
  attn_weight += attn_bias
  attn_weight = torch.softmax(attn_weight, dim=-1)
  if dropout_p > 0:
    attn_weight = torch.dropout(attn_weight, dropout_p, train=True)
  return attn_weight @ value


from jax.sharding import PartitionSpec


def _tpu_flash_attention(query, key, value, env):
  fsdp_partition = PartitionSpec("fsdp")

  def wrap_flash_attention(query, key, value):
    block_sizes = flash_attention.BlockSizes(
        block_b=min(2, query.shape[0]),
        block_q=min(512, query.shape[2]),
        block_k_major=min(512, key.shape[2]),
        block_k=min(512, key.shape[2]),
        block_q_major_dkv=min(512, query.shape[2]),
        block_k_major_dkv=min(512, key.shape[2]),
        block_k_dkv=min(512, key.shape[2]),
        block_q_dkv=min(512, query.shape[2]),
        block_k_major_dq=min(512, key.shape[2]),
        block_k_dq=min(256, key.shape[2]),
        block_q_dq=min(1024, query.shape[2]),
    )
    return flash_attention.flash_attention(
        query, key, value, causal=True, block_sizes=block_sizes)

  if env.config.shmap_flash_attention:
    wrap_flash_attention = shard_map(
        wrap_flash_attention,
        mesh=env._mesh,
        in_specs=(fsdp_partition, fsdp_partition, fsdp_partition),
        out_specs=fsdp_partition,
        check_rep=False,
    )
  # return flash_attn_mapped(query, key, value)
  return wrap_flash_attention(query, key, value)


@register_function(torch.nn.functional.pad)
def pad(tensor, pad, mode="constant", value=None):
  # For padding modes that have different names between Torch and NumPy, this
  # dict provides a Torch-to-NumPy translation. Any string not in this dict will
  # be passed through as-is.
  MODE_NAME_TRANSLATION = {
      "circular": "wrap",
      "replicate": "edge",
  }

  numpy_mode = MODE_NAME_TRANSLATION.get(mode, mode)

  num_prefix_dims = tensor.ndim - len(pad) // 2

  numpy_pad_width = [(0, 0)] * num_prefix_dims
  nd_slice = [slice(None)] * num_prefix_dims

  for i in range(len(pad) - 2, -1, -2):
    pad_start, pad_end = pad[i:i + 2]
    slice_start, slice_end = None, None

    if pad_start < 0:
      slice_start = -pad_start
      pad_start = 0

    if pad_end < 0:
      slice_end = pad_end
      pad_end = 0

    numpy_pad_width.append((pad_start, pad_end))
    nd_slice.append(slice(slice_start, slice_end))

  nd_slice = tuple(nd_slice)

  # `jax.numpy.pad` complains if we provide an irrelevant `constant_values` arg,
  # even if the value we pass in is `None`. (It treats `None` as `nan`.)
  kwargs = dict()
  if mode == "constant" and value is not None:
    kwargs["constant_values"] = value

  # The "replicate" mode pads first and then slices, whereas the "circular" mode
  # slices first and then pads. The latter approach deals with smaller tensors,
  # so we default to that option in modes where the order of operations doesn't
  # affect the result.
  if mode == "replicate":
    return jnp.pad(tensor, numpy_pad_width, mode=numpy_mode, **kwargs)[nd_slice]
  else:
    return jnp.pad(tensor[nd_slice], numpy_pad_width, mode=numpy_mode, **kwargs)


@register_function(
    torch.nn.functional.scaled_dot_product_attention,
    is_jax_function=False,
    needs_env=True,
)
@register_function(
    torch.ops.aten.scaled_dot_product_attention,
    is_jax_function=False,
    needs_env=True)
def scaled_dot_product_attention(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    scale=None,
    enable_gqa=False,
    env=None,
) -> torch.Tensor:

  if env.config.use_tpu_flash_attention:
    jquery, jkey, jvalue = env.t2j_iso((query, key, value))
    res = _tpu_flash_attention(jquery, jkey, jvalue, env)
    return env.j2t_iso(res)

  return _sdpa_reference(query, key, value, attn_mask, dropout_p, is_causal,
                         scale, enable_gqa)


@register_function(
    torch.Tensor.__getitem__, is_jax_function=False, is_view_op=True)
def getitem(self, indexes):

  if isinstance(indexes, list) and isinstance(indexes[0], int):
    # list of int, i.e. x[[1, 2]] NOT x[1, 2] (the second would be tuple of int)
    indexes = (indexes,)
  elif isinstance(indexes, list):
    indexes = tuple(indexes)

  def is_narrow_slicing():
    tensor_free = not pytree.tree_any(
        lambda x: isinstance(x, torch.Tensor) or isinstance(x, jax.Array),
        indexes)
    list_free = not isinstance(indexes, tuple) or all(
        [False if isinstance(x, list) else True for x in indexes])
    return tensor_free and list_free

  if is_narrow_slicing():
    return View(self, view_info=NarrowInfo(indexes), env=self._env)

  indexes = self._env.t2j_iso(indexes)
  return torchax.tensor.Tensor(self._elem[indexes], self._env)


@register_function(torch.corrcoef)
def _corrcoef(x):
  if x.dtype.name == "int64":
    return jnp.corrcoef(x).astype(jnp.float32)
  return jnp.corrcoef(x)


@register_function(torch.sparse.mm, is_jax_function=False)
def _sparse_mm(mat1, mat2, reduce="sum"):
  return torch.mm(mat1, mat2)


@register_function(torch.isclose)
def _aten_isclose(input, other, rtol=1e-05, atol=1e-08, equal_nan=False):
  return jnp.isclose(input, other, rtol, atol, equal_nan)


@register_function(torch.linalg.det)
def linalg_det(input):
  return jnp.linalg.det(input)


@register_function(torch.ones)
def _ones(*size: int, dtype=None, **kwargs):
  if len(size) == 1 and isinstance(size[0], collections.abc.Iterable):
    size = size[0]
  return jaten._ones(size, dtype=dtype)


@register_function(torch.zeros, is_jax_function=True)
def _zeros(*size: int, dtype=None, **kwargs):
  if len(size) == 1 and isinstance(size[0], collections.abc.Iterable):
    size = size[0]
  return jaten._zeros(size, dtype=dtype)


@register_function(torch.eye)
@op_base.convert_dtype()
def _eye(n: int, m: Optional[int] = None, *, dtype=None, **kwargs):
  return jnp.eye(n, m, dtype=dtype)


@register_function(torch.full)
@op_base.convert_dtype()
def _full(size: Sequence[int], fill_value, *, dtype=None, **kwargs):
  # TODO: handle torch.Size
  return jnp.full(size, fill_value, dtype=dtype)


@register_function(torch.empty)
@op_base.convert_dtype()
def empty(*size: Sequence[int], dtype=None, **kwargs):
  if len(size) == 1 and isinstance(size[0], collections.abc.Iterable):
    size = size[0]
  return jnp.empty(size, dtype=dtype)


@register_function(torch.arange, is_jax_function=False)
def arange(
    start,
    end=None,
    step=None,
    out=None,
    dtype=None,
    layout=torch.strided,
    device=None,
    requires_grad=False,
    pin_memory=None,
):
  if end is None:
    end = start
    start = 0
  if step is None:
    step = 1
  return torch.ops.aten.arange(start, end, step, dtype=dtype)


@register_function(torch.empty_strided, is_jax_function=False)
def empty_strided(
    size,
    stride,
    *,
    dtype=None,
    layout=None,
    device=None,
    requires_grad=False,
    pin_memory=False,
):
  return empty(size, dtype=dtype)


@register_function(torch.unravel_index)
def unravel_index(indices, shape):
  return jnp.unravel_index(indices, shape)


@register_function(torch.rand, is_jax_function=False)
def rand(*size, **kwargs):
  if len(size) == 1 and isinstance(size[0], collections.abc.Iterable):
    size = size[0]
  return torch.ops.aten.rand(size, **kwargs)


@register_function(torch.randn, is_jax_function=False)
def randn(
    *size,
    generator=None,
    out=None,
    dtype=None,
    layout=torch.strided,
    device=None,
    requires_grad=False,
    pin_memory=False,
):
  if len(size) == 1 and isinstance(size[0], collections.abc.Iterable):
    size = size[0]
  return torch.ops.aten.randn(size, generator=generator, dtype=dtype)


@register_function(torch.randint, is_jax_function=False)
def randint(*args, **kwargs):
  return torch.ops.aten.randint(*args, **kwargs)


@register_function(torch.logdet)
def logdet(input):
  _, logabsdet = jaten._aten__linalg_slogdet(input)
  return logabsdet


@register_function(torch.linalg.slogdet)
def linalg_slogdet(input):
  sign, logabsdet = jaten._aten__linalg_slogdet(input)
  return torch.return_types.slogdet((sign, logabsdet))


@register_function(torch.tensor_split)
def tensor_split(input, indices_or_sections, dim=0):
  return jnp.array_split(input, indices_or_sections, axis=dim)


@register_function(torch.linalg.solve)
def linalg_solve(a, b):
  res, _ = jaten._aten__linalg_solve_ex(a, b)
  return res


@register_function(torch.linalg.solve_ex)
def linalg_solve_ex(a, b):
  res, info = jaten._aten__linalg_solve_ex(a, b)
  return res, info


@register_function(torch.linalg.svd)
def linalg_svd(a, full_matrices=True):
  return jaten._aten__linalg_svd(a, full_matrices=full_matrices)


@register_function(torch.linalg.matrix_power)
def matrix_power(A, n, *, out=None):
  return jnp.linalg.matrix_power(A, n)


@register_function(torch.svd)
def svd(a, some=True, compute_uv=True):
  if not compute_uv:
    S = jaten._aten__linalg_svd(a, full_matrices=False)[1]
    U = jnp.zeros((a.shape[-2], a.shape[-2]), dtype=a.dtype)
    V = jnp.zeros((a.shape[-1], a.shape[-1]), dtype=a.dtype)
    return U, S, V
  U, S, V = jaten._aten__linalg_svd(a, full_matrices=not some)
  return U, S, jnp.matrix_transpose(V)


@register_function(torch.cdist)
def _cdist(x1, x2, p=2.0, compute_mode="use_mm_for_euclid_dist_if_necessary"):
  return jaten._aten_cdist(x1, x2, p, compute_mode)


@register_function(torch.lu)
def lu(A, **kwargs):
  lu, pivots, _ = jax.lax.linalg.lu(A)
  # JAX pivots are offset by 1 compared to torch
  _pivots = pivots + 1
  info_shape = pivots.shape[:-1]
  info = jnp.zeros(info_shape, dtype=mappings.t2j_dtype(torch.int32))
  if kwargs["get_infos"] == True:
    return lu, _pivots, info
  return lu, _pivots


@register_function(torch.lu_solve)
def lu_solve(b, LU_data, LU_pivots, **kwargs):
  # JAX pivots are offset by 1 compared to torch
  _pivots = LU_pivots - 1
  x = jax.scipy.linalg.lu_solve((LU_data, _pivots), b)
  return x


@register_function(torch.linalg.tensorsolve)
def linalg_tensorsolve(A, b, dims=None):
  # examples:
  # A = torch.randn(2, 3, 6), b = torch.randn(3, 2)
  # A = torch.randn(2, 3, 6), b = torch.randn(2, 3) -> torch.Size([3, 6])
  # A = torch.randn(9, 2, 6, 3) b = torch.randn(6, 3) -> torch.Size([6, 3])
  # A = torch.randn(9, 2, 3, 6) b = torch.randn(6, 3) -> torch.Size([3, 6])
  # A = torch.randn(18, 6, 3) b = torch.randn(18) -> torch.Size([6, 3])
  # A = torch.randn(3, 8, 4, 6) b = torch.randn(4, 6) -> torch.Size([4,6])
  # A = torch.randn(3, 8, 1, 2, 2, 6) b = torch.randn(3, 4, 2) -> torch.Size([2, 2, 6])

  # torch allows b to be shaped differently.
  # especially when axes are moved using dims.
  # ValueError: After moving axes to end, leading shape of a must match shape of b. got a.shape=(3, 2, 6), b.shape=(2, 3)
  # So we are handling the moveaxis and forcing b's shape to match what jax expects
  if dims is not None:
    A = jnp.moveaxis(A, dims, len(dims) * (A.ndim - 1,))
    dims = None
  if A.shape[:b.ndim] != b.shape:
    b = jnp.reshape(b, A.shape[:b.ndim])
  return jnp.linalg.tensorsolve(A, b, axes=dims)


@register_function(torch.nn.functional.linear)
def functional_linear(self, weights, bias=None):
  res = jnp.einsum("...a,ba->...b", self, weights)
  if bias is not None:
    res += bias
  return res


@register_function(torch.nn.functional.interpolate)
def functional_interpolate(
    input,
    size: Tuple[int, int],
    scale_factor: Optional[float],
    mode: str,
    align_corners: bool,
    recompute_scale_factor: bool,
    antialias: bool,
):
  supported_methods = (
      "nearest",
      "linear",
      "bilinear",
      "trilinear",
      "cubic",
      "bicubic",
      "tricubic",
      "lanczos3",
      "lanczos5",
  )
  is_jax_supported = mode in supported_methods
  if not is_jax_supported:
    raise torchax.tensor.OperatorNotFound(
        f"JAX does not support interpolation mode: {mode}. Supported modes are: {supported_methods}"
    )
  # None check
  antialias = antialias or False
  align_corners = align_corners or False

  if mode in ('cubic', 'bicubic',
              'tricubic') and not antialias and size is not None:
    return jimage.interpolate_bicubic_no_aa(
        input,
        size[0],
        size[1],
        align_corners,
    )
  else:
    # fallback
    raise torchax.tensor.OperatorNotFound(
        f"JAX does not support interpolation mode: {mode}. Supported modes are: {supported_methods}"
    )


@register_function(torch.Tensor.repeat_interleave)
def torch_Tensor_repeat_interleave(self,
                                   repeats,
                                   dim=None,
                                   *,
                                   output_size=None):
  return jnp.repeat(self, repeats, axis=dim, total_repeat_length=output_size)
